package de.hawhamburg.reachability.reasoner.engine;

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Random;

import org.apache.log4j.Logger;
import org.drools.KnowledgeBase;
import org.drools.definition.KnowledgePackage;
import org.drools.definition.rule.Global;
import org.drools.runtime.StatefulKnowledgeSession;
import org.drools.runtime.rule.FactHandle;
import org.joda.time.Hours;
import org.joda.time.Minutes;
import org.joda.time.Seconds;

import de.hawhamburg.reachability.ProjectInfo;
import de.hawhamburg.reachability.conf.MainConfig;
import de.hawhamburg.reachability.conf.ReasonerConfig;
import de.hawhamburg.reachability.conf.enumeration.ConfigType;
import de.hawhamburg.reachability.constants.NumericalConstants;
import de.hawhamburg.reachability.enumeration.ReachabilityGroup;
import de.hawhamburg.reachability.enumeration.ReachabilityStatus;
import de.hawhamburg.reachability.json.JsonReachabilitySensorData;
import de.hawhamburg.reachability.reasoner.Reasoner;
import de.hawhamburg.reachability.reasoner.ReasonerAgent;
import de.hawhamburg.reachability.reasoner.engine.drools.DroolsUtil;
import de.hawhamburg.reachability.reasoner.enumeration.ReasonerDimension;
import de.hawhamburg.reachability.sensor.SensorData;
import de.hawhamburg.reachability.util.ExternFileUtil;
import de.hawhamburg.reachability.util.OutputUtil;
import de.hawhamburg.reachability.util.ProjectUtil;
import de.hawhamburg.reachability.util.PropertiesUtil;

/** 
 * A JBoss Drools rule based reasoning engine.<br><br>
 * 
 * This {@link Reasoner} implementation is based on the JBoss Drools rule engine by using several classes of 
 * <code>org.drools</code>. It provides the compilation of author defined XML-Code to drools code by
 * {@link de.hawhamburg.reachability.reasoner.engine.drools.WorkingMemoryRuleGenerator} and is able to handle active 
 * user feedback from user interface. Internal sensor data will be added to the {@link KnowledgeBase} together with the
 * compiled rule set. The rules then will set the current reachability map of this class which will be published onto 
 * the ActiveMQ result topic.
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class DroolsBasedEngine extends Reasoner {
    
    /**
     * The name of the property which contains the name of the current rule-set file.
     */
    private static final String PHENOTYPE_FILE_PROPERTY_KEY = "RULE_PHENOTYPE_FILE";
    
    /**
     * The name of the property which contains the time for how long active feedback shall be minded.
     */
    private static final String MIND_FEEDBACK_KEY           = "REASONING_ENGINE_MIND_FEEDBACK";
    
    /**
     * Basic (user-defined) properties.
     */
    private Properties reasonerProps                        = null;
    
    /**
     * The Drools knowledge base.
     */
    private KnowledgeBase kbase                             = null;
    
    /**
     * The Drools working memory.
     */
    private StatefulKnowledgeSession memory                 = null;
    
    /**
     * Starting time of the reasoner.
     */
    private long startingAt                                 = 0;
    
    /**
     * All handles to facts actual known to Drools working memory.
     */
    private Map<ReasonerDimension, FactHandle> factHandles  =
            new HashMap<ReasonerDimension, FactHandle>();
    
    /**
     * Current given user feedback.
     */
    private Map<ReachabilityGroup, ReachabilityStatus> userFeedback =
            new HashMap<ReachabilityGroup, ReachabilityStatus>();
    
    /**
     * The current status in memory. Important in common the rule base changes in 
     * {@link #updateProperties(Properties, Properties, ConfigType)}.
     */
    private Map<ReachabilityGroup, ReachabilityStatus> memoryStatus =
            new HashMap<ReachabilityGroup, ReachabilityStatus>();
    
    /**
     * The time of the last given user feedback.
     */
    private Map<ReachabilityGroup, Long> lastFeedback       =
            new HashMap<ReachabilityGroup, Long>();
    
    /**
     * The access key which is generated randomly on class creation. It is necessary to know this key to externally 
     * update the reachability status using {@link #setReachabilityStatus(ReachabilityGroup, ReachabilityStatus, long)}.
     */
    private long accessKey                                  = 0;
    
    /**
     * Create a new rule based reasoning engine (Using Jess Rule Engine).
     */
    public DroolsBasedEngine() {
        String propSource   = ReasonerConfig.get("REASONING_ENGINE_PROPERTY_SOURCE");
        
        this.startingAt     = System.currentTimeMillis();
        this.reasonerProps  = PropertiesUtil.getProperties(propSource);
        this.accessKey      = new Random().nextLong();
        
        loadKBase();
    }
    
    /**
     * Load default or cached knowledge base for drools CEP.
     */
    private void loadKBase() {
        // Load the required rule properties
        String phenotypeFile = this.reasonerProps.getProperty("RULE_PHENOTYPE_FILE");
        
        // Create new rule base
        kbase = DroolsUtil.getRules(phenotypeFile);

        // Create a working memory as stateful knowledge session
        if (kbase != null) {
            memory = kbase.newStatefulKnowledgeSession();
            this.setGlobal("REASONER",   this);
            this.setGlobal("ACCESS_KEY", this.accessKey);
        }
    }
    
    /**
     * Sets a global variable of the {@link StatefulKnowledgeSession} to a certain value.
     * 
     * @param name          The name of the global variable within the {@link StatefulKnowledgeSession} as 
     *                      {@link String}.
     * @param value         The new value of the global as {@link Object}.
     */
    private void setGlobal(final String name,
                           final Object value) {
        if (memory != null) {
            if (this.globalExists(name)) {
                memory.setGlobal(name, value);
            } else {
                OutputUtil.error("Required drools global '" + name + "' is undefined within the defined working"
                                 + " memory");
            }
        } else {
            OutputUtil.error("Unable to set drools global '" + name + "' for no working memory is defined");
        }
    }
    
    /**
     * Checks if a global is defined within the {@link StatefulKnowledgeSession} or not.
     * 
     * @param name          The name of the global variable as {@link String}.
     * @return              <code>true</code> if the global is defined, otherwise <code>false</code>.
     */
    private boolean globalExists(final String name) {
        for (KnowledgePackage x : kbase.getKnowledgePackages()) {
            for (Global z : x.getGlobalVariables()) {
                if (z.getName().equals(name)) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * @see de.hawhamburg.reachability.StopableObject#shutDown()
     */
    public final void cleanUp() {
        this.dump();
        
        if (memory != null) {
            memory.dispose();
        }
        
        if (kbase != null) {
            kbase = null;
        }
    }

    /**
     * Adds a new {@link SensorData} object to the {@link StatefulKnowledgeSession} or replaces the last known 
     * {@link SensorData} of this dimension. The {@link ReasonerDimension} and the {@link SensorData} will be extracted 
     * from the {@link JsonReachabilitySensorData} object.
     * 
     * @param sensorData    The {@link JsonReachabilitySensorData} which has been measured.
     */
    public final void setSensorData(final JsonReachabilitySensorData sensorData) {
        String vals = null;
        for (Map.Entry<Integer, Double> x : sensorData.Values.entrySet()) {
            if (vals == null) {
                vals  = x.getKey() + "x" + x.getValue();
            } else {
                vals += ", " + x.getKey() + "x" + x.getValue();
            }
        }
        OutputUtil.info("Receiving new SensorData: " + sensorData.Dimension + " values: " + vals);
        
        // Get data dimension
        ReasonerDimension dimension = sensorData.Dimension;
        
        // Remove old known fact from working memory, if exists
        if (factHandles.containsKey(dimension)) {
            memory.retract(factHandles.get(dimension));
        }
        
        // Add new fact to working memory
        SensorData sd = new SensorData(sensorData.Name, dimension, sensorData.Values);
        factHandles.put(dimension, memory.insert(sd));
        
        // Fire all rules
        memory.fireAllRules();
    }
    
    /**
     * This method will the the {@link ReachabilityStatus} of a certain {@link ReachabilityGroup}. It will only work, 
     * if the correct accessKey is supplied. For this reason, the method can only be called by an internal method or the
     *  drools rule base.
     * 
     * @param group         A {@link ReachabilityGroup}.
     * @param status        A {@link ReachabilityStatus}.
     * @param accessKeyIn   The previously shared access key.
     */
    public final void setReachabilityStatus(final ReachabilityGroup group,
                                            final ReachabilityStatus status,
                                            final long accessKeyIn) {
        if (this.accessKey == accessKeyIn) {
            OutputUtil.info("Set " + group + " to " + status);
            memoryStatus.put(group, status);
        } else {
            OutputUtil.error("Access key '" + accessKeyIn + "' invalid. No rights to set the reachability status!");
        }
    }
    
    /**
     * @see de.hawhamburg.reachability.properties.PropertyServiceNotifiable#updateProperties(Properties, Properties, ConfigType)
     * 
     * @param newProps      The actual defined and correct {@link Properties} set.
     * @param changedProps  The changed {@link Properties} set.
     * @param type          The properties type which shall be updated as {@link ConfigType}.
     */
    public final void updateProperties(final Properties newProps,
                                       final Properties changedProps,
                                       final ConfigType type) {
        switch(type) {
            case REASONER:
                String phenoKey         = PHENOTYPE_FILE_PROPERTY_KEY;
                
                // Assume new properties
                Properties oldProps     = (Properties) this.reasonerProps.clone();
                this.reasonerProps      = newProps;
                
                // Load the required rule properties
                String phenotypeFile    = this.reasonerProps.getProperty(phenoKey);
                
                // Create new rule base
                KnowledgeBase base      = DroolsUtil.getRules(phenotypeFile);
        
                // Create new stateful knowledge session
                if (base != null) {
                    kbase = base;
                    StatefulKnowledgeSession temp = kbase.newStatefulKnowledgeSession();
            
                    // Transfer globals
                    String[] globs = {"REASONER", "ACCESS_KEY"};
                    for (String glob : globs) {
                        Object g = memory.getGlobal(glob);
                        temp.setGlobal(glob, g);
                    }
                    
                    // Transfer facts
                    Map<ReasonerDimension, FactHandle> t = new HashMap<ReasonerDimension, FactHandle>();
                    for (Map.Entry<ReasonerDimension, FactHandle> me : factHandles.entrySet()) {
                        t.put(me.getKey(), temp.insert(memory.getObject(me.getValue())));
                    }
                    factHandles = t;
                    
                    // Memory rewrite
                    memory = temp;
                    
                    // Trigger everything
                    memory.fireAllRules();
                } else {
                    // Reset phenotype file and push error
                    this.reasonerProps.setProperty(phenoKey, oldProps.getProperty(phenoKey));
                    OutputUtil.error("[updateProperties] Couldn't load the new rule base '" + phenotypeFile
                                     + ". Reseted the property '" + phenoKey + "' to '"
                                     + this.reasonerProps.getProperty(phenoKey) + "'");
                    
                }
                
                // Save new properties to file
                String propSource = ReasonerConfig.get(PROPERTY_FILE_SOURCE_KEY);
                PropertiesUtil.save(this.reasonerProps, propSource);
                
                // Notify other systems
                this.pushProperties();
                break;
            case SYSTEM:
                MainConfig.updateProperties(newProps);
                break;
            default:
                break;
        }
        
    }

    /**
     * @see de.hawhamburg.reachability.properties.PropertyServiceNotifiable#loadCurrentProperties(ConfigType)
     * 
     * @param type          Loads the current configuration of a certain type defined as {@link ConfigType}.
     * @return              The requested configuration as {@link Properties} object.
     */
    public final Properties loadCurrentProperties(final ConfigType type) {
        Properties props = new Properties();
        
        switch(type) {
            case REASONER:
              String propSource = ReasonerConfig.get("REASONING_ENGINE_PROPERTY_SOURCE");
              props             = PropertiesUtil.getProperties(propSource);
            default:
                break;
        }
        
        return props;
    }
    
    /**
     * @see de.hawhamburg.reachability.properties.PropertyServiceNotifiable#getValuesForProperty(String)
     * 
     * @param property      The requested property as {@link String}.
     * @return              An empty {@link Map} of {@link String} to {@link String}.
     */
    public final Map<String, String> getValuesForProperty(final String property) {
        return new HashMap<String, String>();
    }

    /**
     * @see de.hawhamburg.reachability.reasoner.Reasoner#getEstimatedReachability()
     * 
     * @return              A {@link Map} of estimated {@link ReachabilityStatus}.
     */
    public final Map<ReachabilityGroup, ReachabilityStatus> getEstimatedReachability() {
        Map<ReachabilityGroup, ReachabilityStatus> result = new HashMap<ReachabilityGroup, ReachabilityStatus>();
        for (ReachabilityGroup p : ReachabilityGroup.values()) {
            result.put(p, getEstimatedReachability(p));
        }
        return result;
    }

    /**
     * @see de.hawhamburg.reachability.reasoner.Reasoner#getEstimatedReachability(ReachabilityGroup)
     * 
     * @param group         The {@link ReachabilityGroup} for which the estimated {@link ReachabilityStatus} shall be 
     *                      returned.
     * @return              The {@link ReachabilityStatus} related to the requested {@link ReachabilityGroup}.
     */
    public final ReachabilityStatus getEstimatedReachability(final ReachabilityGroup group) {
        ReachabilityStatus result = ReachabilityStatus.NOT_CALCULATED;
        
        // Handle known user feedback
        if (lastFeedback.containsKey(group)) {
            long diff    = System.currentTimeMillis() - lastFeedback.get(group);
            String limit = reasonerProps.getProperty(MIND_FEEDBACK_KEY);
            if (diff <= ProjectUtil.convertString(limit, Long.class)) {
                result = userFeedback.get(group);
            } else {
                userFeedback.remove(group);
                lastFeedback.remove(group);
            }
        }
        
        // If no user feedback is given, check if the status was defined by the drools rule engine, based on sensor
        // facts, before.
        if (result.equals(ReachabilityStatus.NOT_CALCULATED)) {
            if (memoryStatus.containsKey(group)) {
                result = memoryStatus.get(group);
            }
        }
        
        return result;
    }

    /**
     * Observes the input topic and handles feedback delivered on this topic.
     * 
     * @param group         The {@link ReachabilityGroup} for which the feedback has been given.
     * @param status        The user expected {@link ReachabilityStatus} to this 
     *                      {@link ReachabilityGroup}.
     */
    public final void handleFeedback(final ReachabilityGroup group,
                                     final ReachabilityStatus status) {
        lastFeedback.put(group, System.currentTimeMillis());
        userFeedback.put(group, status);
    }

    /**
     * @see de.hawhamburg.reachability.reasoner.Reasoner#dump()
     */
    public final void dump() {
        long millsInASecond   = Seconds.ONE.toStandardDuration().getMillis();
        long millsInAMinute   = Minutes.ONE.toStandardDuration().getMillis();
        long millsInAnHour    = Hours.ONE.toStandardDuration().getMillis();
        int  secondsInAMinute = Minutes.ONE.toStandardSeconds().getSeconds();
        
        // Defined context, etc...
        Date date = new Date(this.startingAt);
        long sinceTmp = System.currentTimeMillis() - this.startingAt;
        String since = "";
        if (sinceTmp < 2 * millsInASecond) {
            since = sinceTmp + " milliseconds";
        } else if (sinceTmp / millsInASecond < secondsInAMinute + secondsInAMinute / 2) {
            since = new DecimalFormat("#.###").format(sinceTmp / ((double) millsInASecond)) + " seconds";
        } else if (sinceTmp / millsInAMinute < secondsInAMinute) {
            since = new DecimalFormat("#.###").format(sinceTmp / ((double) millsInAMinute)) + " minutes";
        } else {
            since = new DecimalFormat("#.###").format(sinceTmp / ((double) millsInAnHour)) + " hours";
        }
        
        // Start output
        List<String> info = new ArrayList<String>();
        info.add("");
        info.add("<<<<<<<<<< JBoss Drools Based Reasoning Engine >>>>>>>>>>");
        info.add("");
        info.add("'''''''''''''''''''''''''''''''''''''''''''''''''''''''''");
        info.add("' Engine started at " + date.toString());
        info.add("' Running for " + since);
        info.add("' Wrapper engine by:\t" + ProjectInfo.REACHABILITY_AGENT_AUTHOR);
        info.add("' \t Institute:\t" + ProjectInfo.REACHABILITY_AGENT_INSTITUE);
        info.add("' \t Version:\t" + ProjectInfo.REACHABILITY_AGENT_VERSION);
        info.add("'''''''''''''''''''''''''''''''''''''''''''''''''''''''''");
        
        // Print registered globals
        info.add("' Properties '");
        info.add("''''''''''''''");
        if (this.reasonerProps != null) {
            for (String key : this.reasonerProps.stringPropertyNames()) {
                info.add(key + " = " + this.reasonerProps.getProperty(key));
            }
        } else {
            info.add("!! no properties defined !!");
        }
        info.add("");
        
        // Print facts
        info.add("''''''''''''''''''''");
        info.add("' Registered Facts '");
        info.add("''''''''''''''''''''");
        for (Object f : memory.getObjects()) {
            info.add("<" + f.getClass().getName() + ">\t" + f.toString());
        }
        info.add("");
        
        // Print globals
        info.add("''''''''''''''''''''''");
        info.add("' Registered Globals '");
        info.add("''''''''''''''''''''''");
        String[] globs = {"REASONER", "ACCESS_KEY"};
        for (String glob : globs) {
            Object g = memory.getGlobal(glob);
            if (g != null) {
                info.add("<" + glob + ">\t" + g.toString());
            } else {
                info.add("<" + glob + ">\tundefined");
            }
        }
        info.add("");
        
        // Print globals
        info.add("''''''''''''''''''''''''''''''''''");
        info.add("' Last Known Reachability Status '");
        info.add("''''''''''''''''''''''''''''''''''");
        for (Map.Entry<ReachabilityGroup, ReachabilityStatus> f : memoryStatus.entrySet()) {
            info.add("<" + f.getKey() + ">\t" + f.getValue());
        }
        info.add("");
        
        // Print source
        info.add("'''''''''''''''''''''''''''");
        info.add("' Last Drools Rule source '");
        info.add("'''''''''''''''''''''''''''");
        String[] files = {ReasonerConfig.get("REASONING_ENGINE_DEFAULT_SOURCE"),
                          ReasonerConfig.get("REASONING_ENGINE_RULE_SOURCE") };
        String lnos = "";
        for (String file : files) {
            info.add(file);
            int lno = 0;
            for (String line : ExternFileUtil.getContentLines(file)) {
                lnos = "<" + (++lno) + "> ";
                if (lno < NumericalConstants.TEN) {
                    lnos += " ";
                }
                if (lno < NumericalConstants.HUNDRED) {
                    lnos += " ";
                }
                if (lno < NumericalConstants.THOUSAND) {
                    lnos += " ";
                }
                info.add(lnos + line);
            }
            info.add("");
        }
        
        // Print out all lines...
        for (String line : info) {
            OutputUtil.info(line);
        }
    }
    
    /**
     * Main procedure to start and run this reasoning engine as {@link ReasonerAgent}.
     * 
     * @param args          The system supplied start arguments as array of {@link String}.
     */
    public static void main(final String[] args) {
        // Remove all existing loggers
        Logger.getRootLogger().removeAllAppenders();
        
        // Start this reasoning engine as reasoner agent.
        new ReasonerAgent(new DroolsBasedEngine());
    }
}
