package de.hawhamburg.reachability.reasoner.engine.drools;

import java.io.StringReader;

import org.drools.KnowledgeBase;
import org.drools.KnowledgeBaseConfiguration;
import org.drools.KnowledgeBaseFactory;
import org.drools.builder.KnowledgeBuilder;
import org.drools.builder.KnowledgeBuilderConfiguration;
import org.drools.builder.KnowledgeBuilderFactory;
import org.drools.builder.ResourceType;
import org.drools.io.Resource;
import org.drools.io.ResourceFactory;

import de.hawhamburg.reachability.conf.ReasonerConfig;
import de.hawhamburg.reachability.util.ExternFileUtil;
import de.hawhamburg.reachability.util.OutputUtil;

/**
 * Utility functions related to JBoss Drools.<br><br>
 * 
 * This utility class provides some fancy methods related to the JBoss Drools Framework by JBoss(TM).
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public final class DroolsUtil {
    
    /**
     * Prevent instances of this utility class.
     */
    private DroolsUtil() { }

    /**
     * Creates and loads the currently defined {@link KnowledgeBase} based on the supplied phenotype file {@link String}
     * by using the {@link WorkingMemoryRuleGenerator}.
     * 
     * @param phenotypeFile The path to the phenotype file as {@link String}.
     * @return              The new generated and loaded {@link KnowledgeBase}.
     */
    public static KnowledgeBase getRules(final String phenotypeFile) {
        KnowledgeBase result = null;
        
        if (phenotypeFile.endsWith(".xml") && ExternFileUtil.exists(phenotypeFile)) {
            // Fetch rule source path
            String ruleSource = ReasonerConfig.get("REASONING_ENGINE_RULE_SOURCE");
            
            // Change the rule source
            if (ExternFileUtil.remove(ruleSource)) {
                OutputUtil.info("Start re-creating the rule base.");
                String newRules = new WorkingMemoryRuleGenerator(phenotypeFile).toString();
                ExternFileUtil.createAndWrite(ruleSource, newRules);
                OutputUtil.info("The new rule base was created and written to: " + ruleSource);
            }
            
            result = DroolsUtil.getDefaultKnowledgeBase();
        } else {
            if (phenotypeFile.endsWith(".xml")) {
                OutputUtil.error("[getRules] Unsupported properties file type: " + phenotypeFile);
            } else {
                OutputUtil.error("[getRules] File could not be found: " + phenotypeFile);
            }
        }
        
        return result;
    }
    
    /**
     * Loads and returns the current default {@link KnowledgeBase}. The files <i>REASONING_ENGINE_DEFAULT_SOURCE</i> and
     * <i>REASONING_ENGINE_RULE_SOURCE</i> will be used for creation.
     * 
     * @return              The current default {@link KnowledgeBase}.
     */
    public static KnowledgeBase getDefaultKnowledgeBase() {
        KnowledgeBase kbase = DroolsUtil.getKnowledgeBase();
        
        // Fetch sources
        String defaultSource    = ReasonerConfig.get("REASONING_ENGINE_DEFAULT_SOURCE");
        String ruleSource       = ReasonerConfig.get("REASONING_ENGINE_RULE_SOURCE");
        
        // Load default rules
        if (ExternFileUtil.exists(defaultSource)) {
            KnowledgeBuilder kdefault = DroolsUtil.getKnowledgeBuilderFromFile(defaultSource);
            kbase.addKnowledgePackages(kdefault.getKnowledgePackages());
        } else {
            OutputUtil.error("The required file '" + defaultSource + "' doesn't exist!");
        }
        
        // Load dynamically generated rules
        if (ExternFileUtil.exists(ruleSource)) {
            KnowledgeBuilder krules = DroolsUtil.getKnowledgeBuilderFromFile(ruleSource);
            kbase.addKnowledgePackages(krules.getKnowledgePackages());
        } else {
            OutputUtil.error("The required file '" + ruleSource + "' doesn't exist!");
        }
        
        return kbase;
    }
    
    /**
     * Creates and returns an empty {@link KnowledgeBase}.
     * 
     * @return              An empty {@link KnowledgeBase}.
     */
    public static KnowledgeBase getKnowledgeBase() {
        KnowledgeBaseConfiguration kbconf = KnowledgeBaseFactory.newKnowledgeBaseConfiguration();
        return  KnowledgeBaseFactory.newKnowledgeBase(kbconf);
    }
    
    /**
     * Creates a new {@link KnowledgeBuilder} based on a supplied file by using {@link #getKnowledgeBuilder(Resource)}.
     * 
     * @param file          The file on which the {@link KnowledgeBuilder} shall be based.
     * @return              The {@link KnowledgeBuilder} based on the supplied file.
     */
    public static KnowledgeBuilder getKnowledgeBuilderFromFile(final String file) {
        Resource resource = ResourceFactory.newFileResource(file);
        return DroolsUtil.getKnowledgeBuilder(resource);
    }
    
    /**
     * Creates a new {@link KnowledgeBuilder} based on a supplied rule by using {@link #getKnowledgeBuilder(Resource)}.
     * 
     * @param rule          The rule on which the {@link KnowledgeBuilder} shall be based as {@link String}.
     * @return              The {@link KnowledgeBuilder} based on the supplied rule.
     */
    public static KnowledgeBuilder getKnowledgeBuilderFromString(final String rule) {
        Resource kresource = ResourceFactory.newReaderResource(new StringReader(rule));
        return getKnowledgeBuilder(kresource);
    }
    
    /**
     * Creates an returns a new {@link KnowledgeBuilder} based on a supplied {@link Resource}.
     * 
     * @param resource      The {@link Resource} on which the {@link KnowledgeBuilder} shall be based.
     * @return              The {@link KnowledgeBuilder} based on the supplied {@link Resource}. 
     */
    public static KnowledgeBuilder getKnowledgeBuilder(final Resource resource) {
        KnowledgeBuilderConfiguration kbuconf = KnowledgeBuilderFactory.newKnowledgeBuilderConfiguration();
        
        KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder(kbuconf);
        kbuilder.add(resource, ResourceType.DRL);
        if (kbuilder.hasErrors()) {
            OutputUtil.endSystemWithError(kbuilder.getErrors().toString());
        }
        return kbuilder;
    }
}
