package de.hawhamburg.reachability.reasoner.engine.drools;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import de.hawhamburg.reachability.ProjectInfo;
import de.hawhamburg.reachability.enumeration.ReachabilityGroup;
import de.hawhamburg.reachability.enumeration.ReachabilityStatus;
import de.hawhamburg.reachability.reasoner.enumeration.ReasonerDimension;
import de.hawhamburg.reachability.util.EnumerationUtil;
import de.hawhamburg.reachability.util.OutputUtil;
import de.hawhamburg.reachability.util.ProjectUtil;
import de.hawhamburg.reachability.util.XMLUtil;

/**
 * Class to dynamically create JBoss Drools rule definitions.<br><br>
 * 
 * This class provides methods to generate rule definitions to determine the reachability status of an inhabitant based
 * on user defined configurations.
 *
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class WorkingMemoryRuleGenerator {
    
    /**
     * The package name of the rules base package itself.
     */
    private List<String> packageName = new ArrayList<String>();
    
    /**
     * The path of imports of external java classes and enumerations which shall be included to the rule base.
     */
    private List<String> imports     = new ArrayList<String>();
    
    /**
     * Important globals which shall be included for runtime rule base to java communication.
     */
    private List<String> globals     = new ArrayList<String>();
    
    /**
     * The rules in {@link String} format which will be part of the rule base.
     */
    private List<String> rules       = new ArrayList<String>();
    
    /**
     * This constructor gains the actual defined properties as XML parameter and automatically generates the necessary
     * rule base code in JBoss Drools script language.
     * 
     * <b>Attention:</b> The code will <b>not</b> be returned directly. It will be provided by the {@link #toString()}
     * method. 
     * 
     * @param sourceFile        The path to the source file as {@link String}.
     */
    public WorkingMemoryRuleGenerator(final String sourceFile) {
        
        // Load properties
        Document properties = XMLUtil.loadDocument(sourceFile);
        
        // Meta-Informations
        info("", "", "##########################################");
        info("", "", "### Starting creation of new rule base ###");
        info("", "", "# Sourcefile: '" + sourceFile + "'");
        
        // Set package
        droolsPackageName("rule.base.calcResult");

        // Major class imports required in rules package
        info("", "", "##Start adding imports");
        loadImports(properties);
        
        // Global variables
        info("", "", "##Start writing globals");
        String[][] theGlobals = {
            {"DroolsBasedEngine", "REASONER"},
            {"java.lang.Long",    "ACCESS_KEY"}
        };
        droolsGlobal(theGlobals);
        
        // Define meta rules
        info("", "", "##Start adding loading instructions to rule \"Default\"");
        loadDefaultRule(properties);
        
        // Basic detection rules
        info("", "", "##Start generating rules");
        loadDetectionRules(properties);

        // Creation ended
        info("", "", "### Creation process ended... everything's done ###");
        info("", "", "###################################################");
    }
    
    /**
     * Loads default and user defined import paths to {@link #imports} by using {@link #loadImports(Document)}.
     * 
     * @param properties        The user defined XML file as {@link Document}.
     */
    private void loadImports(final Document properties) {
        String[] defaultImports = {
            "de.hawhamburg.reachability.sensor.SensorData",
            "de.hawhamburg.reachability.reasoner.engine.DroolsBasedEngine",
            "de.hawhamburg.reachability.reasoner.enumeration.ReasonerDimension",
            "de.hawhamburg.reachability.enumeration.ReachabilityGroup",
            "de.hawhamburg.reachability.enumeration.ReachabilityStatus"
        };
            
        String xImports         = "/configuration/imports/import";
        List<String> theImports = new ArrayList<String>(Arrays.asList(defaultImports));
        
        String clazz, pakkage, instruction;
        int c = 0;
        for (Element x : XMLUtil.elementListByXPath(properties, xImports)) {
            instruction = "Import instruction #" + (++c);
            clazz       = x.getTextContent();
            pakkage     = x.getAttribute("package");
            
            // Is attribute 'package' missing or empty?
            if (pakkage.isEmpty()) {
                pakkage = "de.hawhamburg.reachability.sensor.enumeration";
            }
            
            // Is the class known by this project?
            if (ProjectUtil.classExists(pakkage + "." + clazz)) {
                theImports.add(pakkage + "." + clazz);
            } else {
                error("#!" + instruction, xImports, "Unable to import class '" + pakkage + "." + clazz + "'. Reason: "
                                                    + "Class could not be found");
            }
        }
        droolsImport(theImports);
    }

    /**
     * Handles the user defined loading instructions which will be located in the common default rule and which defines
     * variables accessible in all other rules.
     * 
     * @param properties        The user defined XML file as {@link Document}.
     */
    private void loadDefaultRule(final Document properties) {
        String xDeclarations            = "/configuration/declarations/load";
        List<String> declarations       = new ArrayList<String>();
        Map<String, Integer> declared   = new HashMap<String, Integer>();
        
        String variable, dimension, droolsCode, instruction;
        int c = 0;
        for (Element x : XMLUtil.elementListByXPath(properties, xDeclarations)) {
            instruction = "Loading Instruction #" + (++c);
            variable    = x.getAttribute("variable");
            dimension   = x.getAttribute("dimension");
            droolsCode  = x.getTextContent();
            
            // Handling empty or missing attribute 'variable'
            if (variable.isEmpty() && dimension.isEmpty()) {
                error("#!" + instruction, xDeclarations, "Unable to translate instruction. Attributes 'variable' and "
                                                         + "'dimension' are missing or empty.");
            } else if (variable.isEmpty()) {
                variable = "$" + dimension;
                info("#!" + instruction, xDeclarations, "Attribute 'variable' is missing or empty. It has been "
                                                        + "automatically set to" + "'" + variable + "' by dimension "
                                                        + "attribute.");
            }

            // Handling empty, missing or unknown attribute 'dimension'
            if (dimension.isEmpty()) {
                error("#!" + instruction, xDeclarations, "Unable to translate instruction for variable: '" + variable
                                                         + "'.  Attribute 'dimension' is missing or empty.");
            } else if (!EnumerationUtil.contains(ReasonerDimension.class, dimension)) {
                error("#!" + instruction, xDeclarations, "Unable to translate instruction for variable: '" + variable
                                                         + "'. Value '" + dimension + "' of attribute 'dimension' is "
                                                         + "not defined.");
                dimension = "";
            }
            
            // Handling of multiple declared variables
            if (declared.containsKey(variable)) {
                error("#!" + instruction, xDeclarations, "Unable to translate instruction for variable: '" + variable
                                                         + "'. A variable with the same name has been declared before "
                                                         + "by loading instruction #" + declared.get(variable));
                variable = "";
            }
            
            // If variable and dimension are defined, create a new drools loading instruction
            if (!variable.isEmpty() && !dimension.isEmpty()) {
                if (!droolsCode.isEmpty()) {
                    // User defined drools loading code
                    declarations.add(variable + ": " + droolsCode);
                } else {
                    // Default via dimension
                    declarations.add(variable + ": SensorData( dimension == ReasonerDimension." + dimension + " )");
                }
                
                // Memorize that this variable name has been declared
                info("# " + instruction, xDeclarations, "Added loading declaration for " + variable);
                declared.put(variable, c);
            }
        }
        droolsMetaRules(declarations);
    }

    /**
     * Loads the detection rule definitions from user defined xml source and translates them into drools code by using
     * {@link #droolsGroupRule(ReachabilityGroup)} and 
     * {@link #droolsRule(ReachabilityGroup, ReachabilityStatus, String[], boolean)}.
     * 
     * @param properties        The user defined XML file as {@link Document}.
     */
    private void loadDetectionRules(final Document properties) {
        String pBase  = "/configuration/rules";
        boolean fullReplace;
        
        for (ReachabilityGroup group : ReachabilityGroup.values()) {
            // Define XML-Path
            String pGroup = pBase  + "/group[@name='" + group + "']";
            
            // Define meta rules
            droolsGroupRule(group);
            
            for (ReachabilityStatus status : ReachabilityStatus.values()) {
                // Pre-definitions
                List<String> conds = new ArrayList<String>();
                fullReplace = false;
                
                // Get group/state node
                String pStatus = pGroup + "/state[@name='" + status + "']";
                List<Element> defState = XMLUtil.elementListByXPath(properties, pStatus);
                
                // Tag supplied?
                if (defState.size() > 0) {
                    
                    // Evaluate tag "condition"
                    String pCond = pStatus + "/condition";
                    List<Element> defConds = XMLUtil.elementListByXPath(properties, pCond);
                    if (defConds.size() > 0) {
                        
                        // Iterate over all "condition" tags
                        int c = 0;
                        for (Element condition : defConds) {
                            
                            // Counter
                            String pCurCond = pCond + "[" + (++c) + "]";
                            
                            conds.add(getCondition(properties, condition, pCurCond));
                        }
                    }
    
                    // Evaluate tag "code"
                    NodeList codes = defState.get(0).getElementsByTagName("code");
                    if (conds.isEmpty() && codes.getLength() > 0) {
                        List<String> codeLines = getPlainCodeLines(codes, "Rule state node", pStatus + "/code");
                        if (!codeLines.isEmpty()) {
                            conds.addAll(codeLines);
                            fullReplace = true;
                        }
                    }
                    
                    // More than one group/state definition?
                    if (defState.size() > 1) {
                        info("#!Rule state node", pStatus, "Too many definitions with same dimension and state ("
                                                           + defState.size() + "). All but the first ignored!");
                    }
                }
                
                // Nothing done? Then default value..
                if (conds.isEmpty()) {
                    conds.add("eval(false)");
                }
                
                droolsRule(group, status, conds.toArray(new String[conds.size()]), fullReplace);
            }
        }
    }
    
    /**
     * Determines the conditions for a certain user defined rule. The method may call itself recursively and uses the
     * methods {@link #getPlainCodeLine(Element, String, String)} and/or 
     * {@link #getSingleCondition(Element, String, String)}.
     * 
     * @param properties        The user defined XML file as {@link Document}.
     * @param node              The current condition node within the {@link Document} as {@link Element}.
     * @param xPath             The current xPath which leads to the {@link Element} as {@link String}.
     * @return                  The build condition as {@link String} formatted in drools notation.
     */
    private String getCondition(final Document properties,
                                final Element node,
                                final String xPath) {
        String result = "";
        
        // Tertiary attribute of unimatrix zero one
        boolean isNeg = node.hasAttribute("isNegation") && node.getAttribute("isNegation").equals("true");
        boolean code  = node.getNodeName().equals("code");
        
        // Has children?
        int    elmCount     = 0;
        int    c            = 0;
        String levelType    = null;
        String pChilds      = xPath + "/*";
        List<Element> childs = XMLUtil.elementListByXPath(properties, pChilds);
        for (Element child : childs) {
            String type     = child.getNodeName();
            c++;
            
            // First element?
            if (levelType == null) {
                // Check element type
                if (!type.equals("condition") && !type.equals("alternative") && !type.equals("code")) {
                    error("#!" + "Evaluating conditional expression", xPath + "/" + type,
                          "Conditional type '" + type + "' invalid. Expected 'condition', 'alternative' or 'code'");
                    return "";
                }
                
                // Definitions
                levelType = type;
                elmCount  = 1;
                
                // Process node
                result += getCondition(properties, child, xPath + "/" + type + "[" + elmCount + "]");
            } else {
                if (type.equals(levelType)) {
                    // Line prefixing
                    switch(levelType) {
                        case "condition"    : result += " and "; break;
                        case "alternative"  : result += " or "; break;
                        case "code"         : return result;
                        default             : break;
                    }
                    
                    // Counter up
                    elmCount++;

                    // Process node
                    result += getCondition(properties, child, xPath + "/" + type + "[" + elmCount + "]");
                    
                } else {
                    error("#!" + "Evaluating conditional expression", xPath + "/" + type,
                          "Conditional type '" + type + "' unexpected. Only one type will be excepted per level."
                          + " The expected conditional type of this level is '" + levelType + "'");
                }
            }
        }
        
        // No children?
        if (c == 0) {
            // Code element?
            if (node.getTagName().equals("code")) {
                result = getPlainCodeLine(node, "Rule condition", xPath);
            } else {
                result = getSingleCondition(node, "Rule condition", xPath);
            }
        }
        
        // Is expression?
        if (elmCount > 1 && !code) {
            result = "(" + result + ")";
        }
        
        if (isNeg && !code) {
            result = "not " + result;
        }
        
        return result;
    }

    /**
     * Builds a single condition (Tags: <code>alternative</code> or <code>condition</code>) by evaluating the elements
     * text content and attributes. The required attribute is <i>dimension</i>. The attribute <i>value</i> is required
     * if the element has no text content. The default value is <code>eval(false)</code>.
     * 
     * @param node              The current node (<code>alternative</code> or <code>condition</code>) as 
     *                          {@link Element}.
     * @param instruction       A text describing the current instruction for logging reasons as {@link String}.
     * @param xPath             The xPath loading to the current node as {@link String}
     * @return                  The simple condition as {@link String}.
     */
    private String getSingleCondition(final Element node,
                                      final String instruction,
                                      final String xPath) {
        String condition = "";
        
        String dimension = node.getAttribute("dimension");
        String value     = node.getAttribute("value");
        String content   = node.getTextContent();
        
        if (!dimension.isEmpty() && !EnumerationUtil.contains(ReasonerDimension.class, dimension)) {
            error("#!" + instruction, xPath, "Dimension '" + dimension + "' for condition is undefined.");
            dimension = "";
        }
        
        if (!content.isEmpty()) {
            condition = content;
        } else if (!dimension.isEmpty() && !value.isEmpty()) {
            try {
                Double.valueOf(value);
            } catch (NumberFormatException e) {
                value = "(double) " + value + ".ordinal()";
            }
            condition = "SensorData(dimension == ReasonerDimension." + dimension + ", value == " + value + ")";
        } else {
            condition = "eval(false)";
        }
        
        return condition;
    }

    /**
     * Extracts a plain code line from a given user defined code element.
     * 
     * @param code              The user defined code element as {@link Element}.
     * @param instruction       A text describing the current instruction for logging reasons as {@link String}.
     * @param xPath             The xPath leading to the current code element as {@link String}.
     * @return                  The plain code line as {@link String}.
     */
    private String getPlainCodeLine(final Element code,
                                    final String instruction,
                                    final String xPath) {
        String codeLine     = code.getTextContent();
        String result       = null;
        int more            = 1;
        
        for (String line : codeLine.split("\\r?\\n")) {
            if (result == null) {
                result = line;
            } else {
                more++;
            }
        }
        if (more > 1) {
            info("#!" + instruction, xPath, "Too many lines in \"code\" element (" + more + "). All but the first "
                                            + "ignored!");
        }
        
        return result;
    }

    /**
     * Extracts multiple plain code lines from given user defined code elements.
     * 
     * @param codes             The user defined code elements as {@link NodeList}.
     * @param instruction       A text describing the current instruction for logging reasons as {@link String}.
     * @param xPath             The xPath leading to the current code elements as {@link String}.
     * @return                  The plain code lines {@link List} of {@link String}.
     */
    private List<String> getPlainCodeLines(final NodeList codes,
                                           final String instruction,
                                           final String xPath) {
        String codeLine     = codes.item(0).getTextContent();
        List<String> result = new ArrayList<String>();
        
        for (String line : codeLine.split("\\r?\\n")) {
            if (!line.trim().isEmpty()) {
                result.add(line.trim());
            }
        }
        if (codes.getLength() > 1) {
            info("#!" + instruction, xPath, "Too many \"code\" elements (" + codes.getLength() + "). All but the first "
                                            + "ignored!");
        }
        
        return result;
    }
    
    /**
     * Encapsulates the JBoss Drools script code representation of a package name definition. 
     * 
     * @param thePackageName    A {@link String} representing the rule bases package name.
     */
    private void droolsPackageName(final String thePackageName) {
        info("", "", "# Set drools package name to: " + thePackageName);
        packageName.add("package " + thePackageName);
    }
    
    /**
     * Generates imports of multiple paths from the given {@link List} of {@link String} by using 
     * {@link #droolsImport(String)}.
     * 
     * @param clazzes           A {@link List} of {@link String} containing multiple import paths.
     */
    private void droolsImport(final List<String> clazzes) {
        for (String clazz : clazzes) {
            droolsImport(clazz);
        }
    }
    
    /**
     * Encapsulates the JBoss Drools script code representation of an import definition.
     * 
     * @param clazz             A {@link String} representing the path to the class or enumeration which shall be 
     *                          imported.  
     */
    private void droolsImport(final String clazz) {
        info("", "", "# Add class import: '" + clazz + "'");
        imports.add("import " + clazz);
    }

    /**
     * Generates global definitions for multiple key-value pairs from the given two dimensional {@link String} array by 
     * using {@link #droolsGlobal(String, String)}.
     * 
     * @param pairs             A {@link String} array containing multiple key-value pairs for globals.
     */
    private void droolsGlobal(final String[][] pairs) {
        for (String[] pair : pairs) {
            droolsGlobal(pair[0], pair[1]);
        }
    }
    
    /**
     * Encapsulates the JBoss Drools script code representation of an global definition.
     * 
     * @param type              A {@link String} representing the globals type.
     * @param name              A {@link String} representing the name of the global.
     */
    private void droolsGlobal(final String type,
                              final String name) {
        info("", "", "# Add global '" + name + "' of type '" + type + "'");
        globals.add("global " + type + " " + name + ";");
    }
    
    /**
     * Converts a {@link String} array of simple strings to JBoss Drools comment lines and returns them as a 
     * {@link List} of {@link String}.
     * 
     * @param comments          A {@link String} array containing some lines which shall be displayed as comments.
     * @return                  A {@link List} of JBoss Drools formatted comment lines.
     */
    private List<String> droolsComment(final String[] comments) {
        return droolsComment(comments, new ArrayList<String>());
    }

    /**
     * Converts a {@link String} array of simple strings to JBoss Drools comment lines and adds them to a given 
     * {@link List} of code lines.
     * 
     * @param comments          A {@link String} array containing some lines which shall be displayed as comments.
     * @param lines             A {@link List} of code lines to which the comment lines shall be appended.
     * @return                  A {@link List} of JBoss Drools formatted comment lines.
     */
    private List<String> droolsComment(final String[] comments,
                                       final List<String> lines) {
        List<String> result = lines;
        for (String comment : comments) {
            result = droolsComment(comment, lines);
        }
        return result;
    }

    /**
     * Converts a simple {@link String} to a JBoss Drools comment line and adds it to a given {@link List} of code 
     * lines.
     * 
     * @param comment           A {@link String} which shall be displayed as comment.
     * @param lines             A {@link List} of code lines to which the comment line shall be appended.
     * @return                  A {@link List} of JBoss Drools formatted comment lines.
     */
    private List<String> droolsComment(final String comment,
                                       final List<String> lines) {
        List<String> result = lines;
        result.add("// " + comment);
        return result;
    }
    
    /**
     * Adds a definition of a so called "meta" rules by using {@link #droolsMetaRules(String[])} which simply loads 
     * different required variables for rule evaluation.
     * 
     * @param loads             A list of {@link String} representing the different required variables.
     */
    private void droolsMetaRules(final List<String> loads) {
        droolsMetaRules(loads.toArray(new String[loads.size()]));
    }
    
    /**
     * Adds a definition of a so called "meta" rules by using {@link #droolsAddRuleLine(List)} which simply loads 
     * different required variables for rule evaluation.
     * 
     * @param loads             A {@link String} array representing the different required variables.
     */
    private void droolsMetaRules(final String[] loads) {
        String marker = "        ###LOADS###";
        String[] code = {
            "// **************** DEFAULT LOADING RULE *******************************************",
            "rule \"Default\"",
            "    enabled false",
            "    when",
            marker,
            "    then",
            "end",
            ""
        };
        
        // Add loads to marker position
        List<String> codelines = new ArrayList<String>();
        for (String line : code) {
            if (line.equals(marker)) {
                String trailingSpaces = marker.split(marker.trim())[0];
                for (String load : loads) {
                    codelines.add(trailingSpaces + load);
                }
            } else {
                codelines.add(line);
            }
        }
        
        droolsAddRuleLine(codelines);
    }
    
    /**
     * Adds a definition of a reachability groups default rules by using {@link #droolsAddRuleLine(String[])} which 
     * actually does nothing by extending the "meta" rule <i>Default</i>. These rules will by named as combination of 
     * <i>Default</i> and the groups name. It will be extended by the specific rules later.
     * 
     * @param group             A {@link ReachabilityGroup} representing the name of the group for which this default 
     *                          rule shall be defined.
     */
    private void droolsGroupRule(final ReachabilityGroup group) {
        String[] code = {
            "// **************** DETERMINE REACHABILITY FOR " + group.toString().toUpperCase() 
                + " GROUP ***************************",
            "rule \"Default " + group + "\" extends \"Default\"",
            "    enabled false",
            "    when",
            "    then",
            "end",
            ""
        };

        droolsAddRuleLine(code);
    }
    
    /**
     * Adds the certain definition of a rule to determining exact a single {@link ReachabilityStatus} in common with one
     * {@link ReachabilityGroup} by using {@link #droolsAddRuleLine(String[])}. The rule will be named be the 
     * combination of {@link ReachabilityGroup} and {@link ReachabilityStatus} and it extends the default rule of the 
     * related {@link ReachabilityGroup}.<br><br>
     * 
     * The rule actually does nothing by calling a status change method of the reasoner. 
     * 
     * @param group             A {@link ReachabilityGroup} for which the rule shall be created.
     * @param status            A {@link ReachabilityStatus} for which the rule shall be created.
     * @param conditions        A {@link String} array containing the conditions under which this
     *                          {@link ReachabilityStatus} shall be set for the {@link ReachabilityGroup}.
     * @param fullReplace       If set to <code>true</code> the <code>then</code> part of the rule will be replaced
     *                          by the user defined code as well, otherwise not.
     */
    private void droolsRule(final ReachabilityGroup group,
                            final ReachabilityStatus status,
                            final String[] conditions,
                            final boolean fullReplace) {
        info("", "", "# Added rule '" + group + " " + status + "'");
        List<String> code = new ArrayList<String>();
        
        String marker = "        ###CONDITIONS###";
        if (fullReplace) {
            code.add("rule \"" + group + " " + status + "\" extends \"Default " + group + "\"");
            code.add(marker);
            code.add("end");
            code.add("");
        } else {
            code.add("rule \"" + group + " " + status + "\" extends \"Default " + group + "\"");
            code.add("    when");
            code.add(marker);
            code.add("    then");
            code.add("        REASONER.setReachabilityStatus(ReachabilityGroup." + group + ", "
                                                          + "ReachabilityStatus." + status + ", "
                                                          + "ACCESS_KEY);");
            code.add("end");
            code.add("");
        }
        
        // Add loads to marker position
        List<String> codelines = new ArrayList<String>();
        for (String line : code) {
            if (line.equals(marker)) {
                String trailingSpaces = marker.split(marker.trim())[0];
                for (String condition : conditions) {
                    codelines.add(trailingSpaces + condition);
                }
            } else {
                codelines.add(line);
            }
        }

        droolsAddRuleLine(codelines);
    }
    
    /**
     * Adds lines of rule code to the known lines of rule code.
     * 
     * @param code              A {@link String} array containing rule lines of code which shall be registered.
     */
    private void droolsAddRuleLine(final String[] code) {
        for (String line : code) {
            rules.add(line);
        }
    }
    
    /**
     * Adds lines of rule code to the known lines of rule code. 
     * 
     * @param code              A {@link List} of {@link String} containing rule lines of code which shall be 
     *                          registered.
     */
    private void droolsAddRuleLine(final List<String> code) {
        for (String line : code) {
            rules.add(line);
        }
    }
    
    /**
     * Converts a {@link List} of code lines in {@link String} format to a single JBoss Drools code {@link String}, 
     * which is quite easy and nothing but adding a newline (\n) to each line.
     * 
     * @param lines             A {@link List} of {@link String} which shall be converted to a single JBoss Drools line.
     * @return                  A {@link String} as result of the conversion process in JBoss Drools format.
     */
    private String droolsLines(final List<String> lines) {
        String result = "";
        for (String line : lines) {
            result += line + "\n";
        }
        return result;
    }
    
    /**
     * Concatenates JBoss Drools code parts to a single JBoss Drools formatted code {@link String}. 
     * 
     * @param parts             A {@link String} array containing the code parts which shall be concatenated.
     * @return                  A {@link String} containing a single JBoss Drools formatted code line.
     */
    private String droolsConcatParts(final String[] parts) {
        String result = "";
        for (String part : parts) {
            result += part + "\n";
        }
        return result;
    }
    
    /**
     * Handles an info while runtime by using {@link #buildMsg(String, String, String)} and 
     * {@link OutputUtil#info(String)}.
     * 
     * @param instruction       The instruction of the user defined code which was handled as {@link String}.
     * @param xPath             The current xPath within the user defined XML-File as {@link String}.
     * @param message           The info message which shall be published or logged as {@link String}
     */
    private void info(final String instruction,
                      final String xPath,
                      final String message) {
        // Get message
        String infoMsg = buildMsg(instruction, xPath, message);
        
        // Push error
        if (!infoMsg.isEmpty()) {
            OutputUtil.info(infoMsg);
        }
    }
    
    /**
     * Handles an info while runtime by using {@link #buildMsg(String, String, String)} and 
     * {@link OutputUtil#error(String)}.
     * 
     * @param instruction       The instruction of the user defined code which was handled as {@link String}.
     * @param xPath             The current xPath within the user defined XML-File as {@link String}.
     * @param message           The error message which shall be published or logged as {@link String}
     */
    private void error(final String instruction,
                       final String xPath,
                       final String message) {
        // Get message
        String errorMsg = buildMsg(instruction, xPath, message);
        
        // Push error
        if (!errorMsg.isEmpty()) {
            OutputUtil.error(errorMsg);
        }
    }
    
    /**
     * Builds a messages containing the supplied parameters and returns this message as {@link String}.
     * 
     * @param instruction       The instruction of the user defined code which was handled as {@link String}.
     * @param xPath             The current xPath within the user defined XML-File as {@link String}.
     * @param message           The message which shall be published or logged as {@link String}
     * @return                  The resulting message as {@link String}.
     */
    private String buildMsg(final String instruction,
                            final String xPath,
                            final String message) {
        // Build output string
        String msg = "";
        if (instruction != null && !instruction.isEmpty()) {
            msg += instruction.trim() + ": ";
        }
        if (message != null && !message.isEmpty()) {
            msg += message.trim();
            if (!message.trim().endsWith(".")) {
                msg += ". ";
            } else {
                msg += " ";
            }
        }
        if (xPath != null && !xPath.isEmpty()) {
            msg += "Using xpath: '" + xPath + "'.";
        }
        return msg;
    }

    /**
     * Finalizes the output form of the new rule base code {@link String} and returns it.
     * 
     * @return                  The JBoss Drools rule code resulting from the current configuration.
     */
    public final String toString() {
        Date date = new Date(System.currentTimeMillis());
        String[] hcomment = {
            "-----------------------------------------------------------------",
            "Autogenerated JBoss Drools rule base",
            "  Purpose:      Determining the current reachability status",
            "  Generated by: " + ProjectInfo.REACHABILITY_AGENT_NAME,
            "  Version:      " + ProjectInfo.REACHABILITY_AGENT_VERSION,
            "  Author:       " + ProjectInfo.REACHABILITY_AGENT_AUTHOR,
            "  Institute:    " + ProjectInfo.REACHABILITY_AGENT_INSTITUE,
            "  Generated at: " + date.toString(),
            "-----------------------------------------------------------------"
        };
        String headComment = droolsLines(droolsComment(hcomment));
        
        String packageNameArea = droolsLines(this.packageName);
        String importsArea     = droolsLines(this.imports);
        String globalsArea     = droolsLines(this.globals);
        String rulesArea       = droolsLines(this.rules);
        
        String[] parts = {
            headComment,
            packageNameArea,
            importsArea,
            globalsArea,
            rulesArea
        };
        return droolsConcatParts(parts);
    }
    
}
